/*
 * Cantata
 *
 * Copyright (c) 2011-2020 Craig Drummond <craig.p.drummond@gmail.com>
 *
 * ----
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#ifndef CONTEXT_ENGINE_H
#define CONTEXT_ENGINE_H

#include <QObject>
#include <QStringList>

class NetworkJob;

class ContextEngine : public QObject
{
    Q_OBJECT
    
public:
    enum Mode {
        Artist,
        Album,
        Track
    };
    
    static ContextEngine * create(QObject *parent);

    ContextEngine(QObject *p);
    ~ContextEngine() override;
    
    virtual QString translateLinks(QString text) const =0;
    virtual QStringList getLangs() const =0;
    virtual QString getPrefix(const QString &key) const =0;
    QStringList fixQuery(const QStringList &query) const;

    void cancel();

public Q_SLOTS:
    virtual void search(const QStringList &query, Mode mode)=0;
    
Q_SIGNALS:
    void searchResult(const QString &html, const QString &lang);

protected:
    NetworkJob * getReply(QObject *obj);

protected:
    NetworkJob *job;
};

#endif

